/*
#
# Main interface tool for GOPchop.  Implements multiple utility functions.
#
# $Id: main.cpp,v 1.1.1.1 2001/11/16 01:51:46 nemies Exp $
#
# Copyright (C) 2001 Kees Cook
# kees@outflux.net, http://outflux.net/
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# http://www.gnu.org/copyleft/gpl.html
#
*/
/*
 * Initial main.c file generated by Glade. Edit as required.
 * Glade will not overwrite this file.
 */

#include "config.h"

#include <stdio.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#include <string.h>
#include <errno.h>

#include "Vector.h"
#include "List.h"
#include "ElementStream.h"
#include "Pack.h"
#include "GroupOfPictures.h"
#include "Picture.h"
#include "MPEG2Parser.h"

#ifdef __cplusplus
extern "C" {
#endif
#include <mpeg2dec/video_out.h>
#include <mpeg2dec/mpeg2.h>
#include <mpeg2dec/mm_accel.h>
#ifdef __cplusplus
	   }
#endif

#include <gtk/gtk.h>
#include <math.h>

#ifdef __cplusplus
extern "C" {
#endif
#include "interface.h"
#include "support.h"
#ifdef __cplusplus
           }
#endif

#include "main.h"

#include <getopt.h>



/* */
/* globals for everyone to beat on */
/* */
// strings
char * GOPchop_cvsid="$Id: main.cpp,v 1.1.1.1 2001/11/16 01:51:46 nemies Exp $";
char * GOPchop_version=VERSION;

// windows
GtkWidget *main_window=NULL;
GtkWidget *about_window=NULL;
GtkWidget *fileselection_window=NULL;
GtkWidget *saveselection_window=NULL;
GtkWidget *error_dialog=NULL;
GtkWidget *overwrite_dialog=NULL;
GtkWidget *GOP_window=NULL;

// the menus and buttons
GtkWidget * menu_open=NULL;
GtkWidget * menu_save=NULL;
GtkWidget * menu_close=NULL;
GtkWidget * button_run=NULL;
GtkWidget * button_refresh=NULL;
GtkWidget * button_prev=NULL;
GtkWidget * button_next=NULL;
GtkWidget * button_start_mark=NULL;
GtkWidget * button_end_mark=NULL;
GtkWidget * menu_options_run_loop=NULL;
GtkWidget * menu_options_auto_refresh=NULL;
GtkWidget * main_popup=NULL;
popup_info_t main_popup_info;
gint	    main_clist_row=-1;

// labels and things
GtkWidget * main_progressbar=NULL;
GtkWidget * main_statusbar=NULL;
GtkWidget * main_clist=NULL;
GtkWidget * main_label_mark=NULL;
GtkWidget * error_text_why=NULL;
GtkWidget * GOP_selector=NULL;
GtkWidget * GOP_label_filename=NULL;
GtkWidget * overwrite_label_filename=NULL;
GtkWidget * GOP_clist=NULL;


// decode variables
static mpeg2dec_t mpeg2dec;
static vo_open_t * output_open=NULL;
static vo_instance_t * output=NULL;


///// Callbacks that glade doesn't know about
void
on_GOP_selector_value_changed           (GtkAdjustment *adjustment,
                                         gpointer user_data)
{
	int GOP;

	GOP=(int)GTK_ADJUSTMENT(adjustment)->value;

	// only do something if we have output, and auto refreshing
	if (output && GTK_CHECK_MENU_ITEM(menu_options_auto_refresh)->active)
		show_GOP(GOP);

	update_GOP_info(GOP);
}



///// My various supporting functions
void
flush()
{
	while (g_main_iteration(FALSE));
	/*
        while(gtk_events_pending())
                gtk_main_iteration();
	*/
}

void
status_on(const gchar * owner, char * text)
{
        guint context;

	context=gtk_statusbar_get_context_id(GTK_STATUSBAR(main_statusbar),
			owner);
        gtk_statusbar_push(GTK_STATUSBAR(main_statusbar), context, text);
        flush();
}

void
status_off(const gchar * owner)
{
        guint context;

	context=gtk_statusbar_get_context_id(GTK_STATUSBAR(main_statusbar),
			owner);
        gtk_statusbar_pop(GTK_STATUSBAR(main_statusbar), context);
        flush();
        //gtk_main_iteration_do(0);

}       

const gchar * fileops="fileops";

void
progress(char * task, float done)
{
        static gchar *  lasttask=NULL;
        static float    lastdone=0.0;
        float diff;

        int tasksdiffer;
        gint context;

	context=gtk_statusbar_get_context_id(GTK_STATUSBAR(main_statusbar),
			fileops);

        if (lasttask)
        {
                tasksdiffer=strcmp(lasttask,task);
        }
        else
        {
                tasksdiffer=1;
        }

        if (tasksdiffer)
        {
                if (lasttask)
                        gtk_statusbar_pop(GTK_STATUSBAR(main_statusbar),
					context);

                lasttask=(char*)g_realloc(lasttask,strlen(task)+1);
                strcpy(lasttask,task);

                gtk_statusbar_push(GTK_STATUSBAR(main_statusbar), context,
				lasttask);
                flush();
        }

        diff=fabs(done-lastdone);

        if (!tasksdiffer && diff<0.01)
                return;

        lastdone=done;

	gtk_progress_set_percentage(GTK_PROGRESS(main_progressbar),lastdone);
        flush();
}

void
show_error(char * text)
{
  static gchar * str=NULL;

  // text needs a static area, I think... ?
  (void*)str=g_realloc(str,strlen(text)+1);
  strcpy(str,text);

  gtk_text_freeze(GTK_TEXT(error_text_why));
  gtk_text_set_point(GTK_TEXT(error_text_why),0);    
  gtk_text_forward_delete(GTK_TEXT(error_text_why),
		  gtk_text_get_length(GTK_TEXT(error_text_why)));
  gtk_text_insert(GTK_TEXT(error_text_why), NULL, NULL, NULL,
		  (str), strlen(str));
  gtk_text_set_point(GTK_TEXT(error_text_why),
		  gtk_text_get_length(GTK_TEXT(error_text_why)));     
  gtk_text_thaw(GTK_TEXT(error_text_why));

  gtk_widget_show(GTK_WIDGET(error_dialog));
}

void
write_GOP(FILE * file, int num, int continuous)
{
	List * GOPs;
	GroupOfPictures * GOP;
	List * packets;
	Bounds * packet_bounds;
	Vector * pes_vector;
	Pack *	packet;
	int pes,pes_max;
	uint8_t old_bits;
	uint8_t * header;
	uint8_t * area;

	// FIXME: disabled broken link setting...
	continuous=1;

	if (!(GOPs=mpeg2parser->getGOPs()))
	{
		printf("NULL GOP list?!\n");
		return;
	}
	if (!(packets=mpeg2parser->getPackets()))
	{
		printf("NULL packet list?!\n");
		return;
	}
	if (!(GOP=(GroupOfPictures*)GOPs->vector(num)))
	{
		printf("NULL GOP?!\n");
		return;
	}
	if (!(packet_bounds=GOP->getPacketBounds()))
	{
		printf("NULL packet bounds?!\n");
		return;
	}

	// handle broken flag
	if (!continuous)
	{
		if (!(header=mpeg2parser->bytesAvail(GOP->getStart(),
						GOP->getLen())))
		{
			printf("unavailable memory for GOP header?!\n");
			return;
		}
		old_bits=header[7];
		printf("broken was %d\n",(old_bits&GOP_FLAG_BROKEN)>0);
		header[7]|=GOP_FLAG_BROKEN;
		printf("broken is now %d\n",(header[7]&GOP_FLAG_BROKEN)>0);
	}

	pes_max=packet_bounds->getMax();
	for (pes=packet_bounds->getFirst();pes<pes_max;pes++)
	{
		size_t	len;

		if (!(packet=(Pack*)packets->vector(pes)))
		{
			printf("NULL PES?!\n");
			goto abort;
		}

		len=packet->getLen();
		if (!(area=mpeg2parser->bytesAvail(packet->getStart(),len)))
		{
			printf("unavailable memory?!\n");
			goto abort;
		}

		if (fwrite(area,sizeof(uint8_t),len,file)!=len)
		{
			printf("write failed\n");
			goto abort;
		}
	}
abort:
	if (!continuous)
		header[7]=old_bits;
}

void
save_file(char * filename)
{
	char buf[1024];
	FILE * mpeg2;
	char * task="Saving MPEG2";
	uint8_t end_buf[4]={0x0, 0x0, 0x1, 0xB9};

	if (!(mpeg2=fopen(filename,"w")))
	{
		snprintf(buf,1024,"%s: %s\n",filename,strerror(errno));
		show_error(buf);
	}
	else
	{
		gint row;
		int i;
		gchar *starts;
		gchar *ends;
		int start, end;
		int clistrows;
		int totalGOPs=0;
		int count;
		int prev_GOP=-1;

		for (row=0;
		     gtk_clist_get_text(GTK_CLIST(main_clist),row,1,&starts);
		     row++)
		{ }
		clistrows=row;

		for (row=0;
		     row<clistrows;
		     row++)
		{
		        gtk_clist_get_text(GTK_CLIST(main_clist),row,1,&starts);
		     	gtk_clist_get_text(GTK_CLIST(main_clist),row,2,&ends);

			start=atoi(starts);
			end=atoi(ends);

			totalGOPs+=end-start+1;
		}

		if (!totalGOPs)
			goto abort;

		count=0;
		for (row=0;
		     row<clistrows;
		     row++)
		{
		        gtk_clist_get_text(GTK_CLIST(main_clist),row,1,&starts);
		     	gtk_clist_get_text(GTK_CLIST(main_clist),row,2,&ends);

			start=atoi(starts);
			end=atoi(ends);

			for (i=start;i<=end;i++)
			{
				progress(task,
					(float)((float)count++/(float)totalGOPs));

				write_GOP(mpeg2,i,(prev_GOP+1)==i);
				prev_GOP=i;
			}
		}
		progress(task,1.0);
		status_off(fileops);
abort:
		//fwrite(end_buf,sizeof(uint8_t),4,mpeg2);
		fclose(mpeg2);
	}
}

void
open_file(char * filename)
{
        if (!mpeg2parser->init(filename,progress))
        {
                show_error(mpeg2parser->getError());
        }
        else {
                mpeg2parser->parse();
                if (mpeg2parser->getError())
                        show_error(mpeg2parser->getError());

                // make sure we actually HAVE something...
                List * GOPs=mpeg2parser->getGOPs();
                if (GOPs && GOPs->getNum())
                        file_is_loaded();
        }
        progress("Done",1.0);
        status_off(fileops);
}

void
add_GOP_slice(int start, int end)
{
	gchar * pathname;
	gchar first[128];
	gchar last[128];
	gchar * data[3];

        pathname=mpeg2parser->getFilename();
        if (!pathname)
                pathname="unknown";

	data[0]=pathname;
	data[1]=first;
	data[2]=last;

	sprintf(first,"%d",start);
	sprintf(last,"%d",end);

	gtk_clist_append(GTK_CLIST(main_clist),data);
}


// decode functions
void
decode_stop()
{
	printf("decode stop\n");

	if (output)
	{
		mpeg2_close(&mpeg2dec);
		vo_close(output);
		output=NULL;
		output_open=NULL;
		memset(&mpeg2dec,0,sizeof(mpeg2dec));
	}
}

void
decode_start()
{
        int i;
#ifdef BROKEN_SPLASH
	int fd;
	struct stat st;
	size_t filesize;
	uint8_t * splash;
#endif

        // decoding
        uint32_t accel=0;
        vo_driver_t * drivers;

	printf("decode start\n");

        drivers=vo_drivers();
        for (i=0;drivers[i].name;i++)
                if (!strcmp(drivers[i].name,"x11"))
                        output_open=drivers[i].open;

        if (!output_open)
        {
                fprintf(stderr,"can't find 'x11' output\n");
                exit(1);
        }

        accel=(mm_accel () | MM_ACCEL_MLIB);
        vo_accel(accel);
        output=vo_open(output_open);

        if (!output)
        {
                fprintf(stderr,"can't open 'x11' output\n");
                exit(1);
        }

        mpeg2_init(&mpeg2dec, accel, output);

#ifdef BROKEN_SPLASH
        // open file
        if ((fd=open("splash.m2v",O_RDONLY))<0)
        {
		perror("splash.m2v");
		exit(1);
        }

        // get file size
        if (fstat(fd,&st))
        {
		perror("fstat");
		exit(1);
        }

        if (!S_ISREG(st.st_mode))
        {
		fprintf(stderr,"not a regular file\n");
		exit(1);
        }

	filesize=st.st_size;

        if ((splash=(uint8_t *)mmap(NULL,filesize,PROT_READ,
                                         MAP_SHARED,fd,0)) == MAP_FAILED)
        {
		perror("mmap");
		exit(1);
        }

	// get window up on the screen
	mpeg2_decode_data(&mpeg2dec, splash, splash+filesize);

        if (munmap(splash,filesize))
        {
		perror("munmap");
		exit(1);
        }
	close(fd);
#endif
}

void
update_GOP_info(int num)
{
	int i;
	GroupOfPictures * GOP;
	Bounds * picture_bounds;
	Bounds * packet_bounds;
	Picture * picture;
	List *	GOPs;
	List *	picture_list;
	List *	packet_list;
	List *	audio_list;
	List *	video_list;
	Vector * head_vector;
	int	picture_index, picture_max;
	int	packet_index, packet_max;
	//int	aindex, amax;
	uint8_t*head;
	int	drop, hour, min, sec, time;

	gchar datatype[128];
	gchar dataoffset[128];
	gchar datasize[128];
	gchar *data[3]={ datatype, dataoffset, datasize };

	gtk_clist_freeze(GTK_CLIST(GOP_clist));

	gtk_clist_clear(GTK_CLIST(GOP_clist));

	if (!(GOPs=mpeg2parser->getGOPs()))
	{
		printf("no GOPs ?!\n");
		goto failure;	
	}
	if (!(packet_list=mpeg2parser->getPackets()))
	{
		printf("NULL picture list ?!\n");
		goto failure;	
	}
	if (!(picture_list=mpeg2parser->getPictures()))
	{
		printf("NULL picture list ?!\n");
		goto failure;	
	}
	if (!(audio_list=mpeg2parser->getAudio()))
	{
		printf("NULL audio list ?!\n");
		goto failure;	
	}
	if (!(video_list=mpeg2parser->getVideo()))
	{
		printf("NULL video list ?!\n");
		goto failure;	
	}
	if (!(GOP=(GroupOfPictures*)GOPs->vector(num)))
	{
		printf("no such GOP %d ?!\n",num);
		goto failure;
	}
	if (!(head_vector=GOP->getHeader()))
	{
		printf("GOP header missing?!\n");
		goto failure;
	}
	if (!(head=mpeg2parser->bytesAvail(head_vector->getStart(),8)))
	{
		printf("GOP header no available?!\n");
		goto failure;
	}

	/*
	    4         5         6
                 |         |         |
	7 65432 107654 3 210876 543210
	1 11111 111111 1 111111 111111
	d hour  min    m sec    pic
	*/
	drop=((head[4]&0x80)>0);
	hour=((head[4]&0x7C)>>2);
	min=((head[4]&0x3)<<4)|((head[5]&0xF0)>>4);
	sec=((head[5]&0x7)<<3)|((head[6]&0xE0)>>6);
	time=(head[6]&0x3F);

	sprintf(datatype,"GOP (%02d:%02d:%02d - %02d %sdrop)",
			hour,min,sec,time,drop?"":"no ");
	sprintf(dataoffset,"%d",num);
	sprintf(datasize,"N/A");
	gtk_clist_append(GTK_CLIST(GOP_clist),data);

	if (!(picture_bounds=GOP->getPictureBounds()))
	{
		printf("NULL picture bounds?!\n");
		goto failure;
	}

	// FIXME: technically, this loop does a lot of needless
	//        reassigning of values.  However, this loop doesn't
	//        happen often, and it's the easiest way I could think
	//        of to sort by offset.

	// get picture bounds
	picture_index=picture_bounds->getFirst();
	picture_max=picture_bounds->getMax();

	for (;picture_index<picture_max;picture_index++)
	{
		if (!(picture=(Picture*)picture_list->vector(picture_index)))
		{
			printf("NULL picture?!\n");
			goto failure;
		}

		sprintf(datatype,"  Picture (%s: %02d)",
				frame_type_str(picture->getType()),
				picture->getTime());
		sprintf(dataoffset,"%d",picture_index);
		sprintf(datasize,"N/A");
		gtk_clist_append(GTK_CLIST(GOP_clist),data);
	}

#if 0 // FIXME: pictures seem to be totally fux0r'd
	// sort PES chunks by offset
	while (pindex<pmax || aindex<amax)
	{
		Vector * vbounds;
		int 	vfirst, vindex, vmax;

		if (!(picture=(Picture*)plist->vector(pindex)))
		{
			printf("NULL picutre?!\n");
			goto failure;
		}
		if (!(vbounds=picture->getVideoBounds()))
		{
			printf("NULL video bounds?!\n");
			goto failure;
		}

		// get video bounds
		vfirst=vindex=vbounds->getStart();
		vmax=vindex+vbounds->getLen();

		while (aindex<amax || vindex<vmax)
		{
			ElementStream * ves;
			ElementStream * aes;
			off_t	vloc, aloc;

			if (!(ves=(ElementStream*)vlist->vector(vindex)))
			{
				printf("NULL video?!\n");
				goto failure;
			}
			if (!(aes=(ElementStream*)alist->vector(aindex)))
			{
				printf("NULL audio?!\n");
				goto failure;
			}

			vloc=ves->getStart();
			aloc=aes->getStart();

			if (vloc<aloc)
			{
				// show video ES first
				sprintf(datatype,"  Video");
				// report Picture frame type
				if (vindex==vfirst)
				{
					sprintf(datatype+strlen(datatype),
							" (%s)",
							frame_type_str(picture->getType()));
				}
				sprintf(dataoffset,"%lld",vloc);
				sprintf(datasize,"%ld",ves->getLen());
				vindex++;
			}
			else
			{
				// show audio ES first
				sprintf(datatype,"  Audio");
				sprintf(dataoffset,"%lld",aloc);
				sprintf(datasize,"%ld",aes->getLen());
				aindex++;
			}
			gtk_clist_append(GTK_CLIST(GOP_clist),data);
		}
		if (vindex==vmax)
			pindex++;
	}
#endif

failure:
	gtk_clist_thaw(GTK_CLIST(GOP_clist));
}

char * frame_type_map[4]={"Bad", "I-Frame", "P-Frame", "B-Frame"};
char *
frame_type_str(int type)
{
	if (type>3 || type<0)
		type=0;
	return frame_type_map[type];
}


void
show_GOP(int wantedGOP)
{
	static int initdone=0;
	List * packets;
	Pack * packet;
	List * GOPs;
	GroupOfPictures * GOP;
       	Bounds * bounds;
       	ElementStream * ves;
       	int start, stop;
       	int ves_start, ves_stop;

       	uint8_t * loc;
       	size_t len;

	// decoder isn't running...
	if (!output)
	{
		printf("\tdecoder not running?\n");
		return;
	}

	// get our GOP list
	if (!(GOPs=mpeg2parser->getGOPs()))
	{
		printf("\tNo GOPs\n");
		return;
	}

	// get our packet list
	if (!(packets=mpeg2parser->getPackets()))
	{
		printf("\tNo packets\n");
		return;
	}

	// get GOP we're after
       	if (!(GOP=(GroupOfPictures*)GOPs->vector(wantedGOP)))
	{
		printf("\tGOP missing\n",wantedGOP);
		return;
	}

#if 0
#ifndef BROKEN_SPLASH
	// for some reason, we have to do this after our
	// GTK window initializations??  FIXME
	if (!initdone)
	{
        	decode_init();
		initdone=1;
	}
#endif
#endif

	// get our GOP's range of packets
       	bounds=GOP->getPacketBounds();
	// remember our start and finish points
       	start=bounds->getFirst();
       	stop=bounds->getMax();

       	for (int j=start;j<stop;j++)
       	{
		if (!(packet=(Pack*)packets->vector(j)))
		{
			printf("NULL packet?!\n");
			exit(1);
		}

		ves_start=packet->getVideoFirst();
		ves_stop=packet->getVideoMax();

		for (int k=ves_start;k<ves_stop;k++)
		{
			if (!(ves=(ElementStream*)mpeg2parser->getVideo()->vector(k)))
			{
				printf("NULL ves?!\n");
				exit(1);
			}

			len=ves->getLen();
			(void*)loc=mpeg2parser->bytesAvail(ves->getStart(),len);

			/*
			fprintf(stderr,"\t\tVES: %d @ %lld (%d): 0x%08x\n",
					k,ves->getStart(),len,
					(unsigned int)loc);
			*/
                                
			if (loc && len)
				mpeg2_decode_data(&mpeg2dec, loc, loc+len);
		}
	}
}

int
get_GOP_selected()
{
	int GOP;

	GOP=(int)GTK_ADJUSTMENT(gtk_range_get_adjustment(GTK_RANGE(GOP_selector)))->value;
	return GOP;
}

void
set_GOP_selected(int num)
{
	int max;
	int old;

	max=mpeg2parser->numGOPs();
	// correct our bounds
	if (num>=max)
		num=max-1;
	if (num<0)
		num=0;

	// remember where we were
	old=get_GOP_selected();

	gtk_adjustment_set_value(GTK_ADJUSTMENT(gtk_range_get_adjustment(GTK_RANGE(GOP_selector))),num);

	// if we're in the same place, force a refresh, damnit
	if (num==old)
		gtk_adjustment_value_changed((GTK_ADJUSTMENT(gtk_range_get_adjustment(GTK_RANGE(GOP_selector)))));
}

void
Usage(char * title)
{
	printf("Usage: %s [OPTIONS] [FILE]\n",title);
	printf("\n");
	printf("  -h, --help         This help\n");
	printf("  FILE               MPEG2 to load on startup\n");
	printf("\n");

	exit(1);
}


// objects
MPEG2Parser * mpeg2parser;

int
main (int argc, char *argv[])
{
  const gchar * mainstatus="main";
  GtkAdjustment * scale;
  int c;
      
  static struct option long_options[] = {
	{ "help", 0, NULL, 'h' }, // 0
	{ 0, 0, 0, 0 }
  };    


#ifdef ENABLE_NLS
  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  textdomain (PACKAGE);
#endif

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  // do arg handling
  while (1) {
        int index;
        c = getopt_long(argc, argv, "h", long_options, &index);
        
        if (c==-1) break; // done with args
        
        switch (c) {
        case 0:
                // got a matching long argument (with no short equiv)
                switch (index) {
                default:
                        fprintf(stderr,"Unknown long argment index %d found!\n",
                                index);
                }               
                break;
        case 'h':
                Usage(argv[0]);
                break;
        default:
                fprintf(stderr,"Unknown short argment '%c' found?!\n",c);
        case '?':
                Usage(argv[0]);
        }
        break;
  }

  add_pixmap_directory (PACKAGE_DATA_DIR "/pixmaps");
  add_pixmap_directory (PACKAGE_SOURCE_DIR "/pixmaps");
  

  mpeg2parser=new MPEG2Parser;

  /*
   * The following code was added by Glade to create one of each component
   * (except popup menus), just so that you see something after building
   * the project. Delete any components that you don't want shown initially.
   */
  main_window = create_main_window ();

  // create our error dialog now
  error_dialog = create_error_dialog ();
  overwrite_dialog = create_overwrite_dialog ();

  // create our GOP window now
  GOP_window = create_GOP_window ();

  // create our popup
  main_popup = create_main_popup ();
  main_popup_info.x=0;
  main_popup_info.y=0;

  // locate menus and buttons
  menu_open=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "menu_file_open"));
  menu_close=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "menu_file_close"));
  menu_save=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "menu_file_save"));
  menu_options_run_loop=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "menu_options_run_loop"));
  menu_options_auto_refresh=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "menu_options_auto_refresh"));
  button_run=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "button_run"));
  button_refresh=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "button_refresh"));
  button_prev=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "button_prev"));
  button_next=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "button_next"));
  button_start_mark=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "button_start_mark"));
  button_end_mark=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "button_end_mark"));

  // locate things
  main_progressbar=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "main_progressbar"));
  main_statusbar=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "main_statusbar"));
  GOP_selector=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "GOP_selector"));
  main_label_mark=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "main_label_mark"));
  main_clist=GTK_WIDGET(lookup_widget (GTK_WIDGET(main_window), "main_clist"));
  main_clist_row=-1; // nothing selected
  GOP_clist=GTK_WIDGET(lookup_widget (GTK_WIDGET(GOP_window), "GOP_clist"));
  GOP_label_filename=GTK_WIDGET(lookup_widget (GTK_WIDGET(GOP_window), "GOP_label_filename"));
  overwrite_label_filename=GTK_WIDGET(lookup_widget (GTK_WIDGET(overwrite_dialog), "overwrite_label_filename"));
  error_text_why=GTK_WIDGET(lookup_widget (GTK_WIDGET(error_dialog), "error_text_why"));

  ////////////////////
  // do things that glade doesn't handle
  ///////////////////
  // set up ranges
  scale=gtk_range_get_adjustment(GTK_RANGE(GOP_selector));
  scale->step_increment=1;
  scale->page_increment=1;
  scale->page_size=1;
  gtk_signal_connect(GTK_OBJECT (scale), "value-changed",
		     GTK_SIGNAL_FUNC(on_GOP_selector_value_changed),
		     NULL);

  // column justification
  gtk_clist_set_column_justification (GTK_CLIST (main_clist), 1,
		  GTK_JUSTIFY_RIGHT);
  gtk_clist_set_column_justification (GTK_CLIST (main_clist), 2,
		  GTK_JUSTIFY_RIGHT);
  gtk_clist_set_column_justification (GTK_CLIST (GOP_clist), 1,
		  GTK_JUSTIFY_RIGHT);
  gtk_clist_set_column_justification (GTK_CLIST (GOP_clist), 2,
		  GTK_JUSTIFY_RIGHT);

  // clist options
  gtk_clist_set_column_min_width (GTK_CLIST (main_clist), 0, 200);
  gtk_clist_set_column_auto_resize (GTK_CLIST (main_clist), 0, TRUE);
  gtk_clist_set_column_auto_resize (GTK_CLIST (main_clist), 1, TRUE);
  gtk_clist_set_column_auto_resize (GTK_CLIST (main_clist), 2, TRUE);

  gtk_clist_set_reorderable(GTK_CLIST (main_clist), TRUE);
  gtk_clist_set_use_drag_icons(GTK_CLIST (main_clist), FALSE);





  // word wrap
  gtk_text_set_word_wrap(GTK_TEXT(error_text_why), TRUE);

  // further initialization
  status_on(mainstatus, "Ready");


  // verify that we're working with largefile support
  if (sizeof(off_t)<8)
  {
	  show_error("Type 'off_t' is less than 8 bytes.\nLargefile support requires at least 8 bytes.\n");
	  gtk_widget_show(error_dialog);
  }

  gtk_widget_show (main_window);
  gtk_widget_show (GOP_window);

  if (optind<argc)
	  open_file(argv[optind]);

  gtk_main ();
  return 0;
}

